//
// C++ Interface: kpgdatatable
//
// Description: 
//
//
// Author: Lumir Vanek <lvanek@users.sourceforge.net>, (C) 2006
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef KPGDATATABLE_H
#define KPGDATATABLE_H

// Note that for new code which you wish to be strictly Unicode-clean, you can define the macro QT_NO_ASCII_CAST when compiling your code.
#define QT_NO_ASCII_CAST

// include files for Qt
#include <qtable.h>
#include <qvaluelist.h>

// include files for KDE
#include <kurl.h>

// application specific includes
#include "../kpgconnectioninthread.h"
#include "../DbObjects/kpgtableindex.h"
#include "../kpgactionsstates.h"

// include files for libpgxx 
#include <pqxx/largeobject.hxx>

class QTextCodec;
class QTable;
class KFind;
class KPGConnection;
class KPGTable;
class KPGDatabase;


/**
* Stores information about column in data table
*
*/
class KPGDataTableColumn
{
public:
	KPGDataTableColumn() {};
	KPGDataTableColumn(const QString &strName, const QString &strTypName, bool bNotNull, bool bHasDef, const QString &strAdSrc)
	{
		m_strName = strName;
		m_strTypName = strTypName;
		m_bNotNull = bNotNull;
		m_bHasDef = bHasDef;
		m_strAdSrc = strAdSrc;
	}
	
	const QString &name() const { return m_strName; }
	const QString &typName() const { return m_strTypName; }
	bool hasDefault() const { return m_bHasDef; }
	bool isNotNull() const { return m_bNotNull; }
	const QString &defaultValue() const { return m_strAdSrc; }
	
protected:
    QString m_strName;
    QString m_strTypName;
	bool m_bNotNull;
	bool m_bHasDef;
	QString m_strAdSrc;
};

/**
* Stores information about field in data table
*
*/
class KPGDataTableField
{
public:
	KPGDataTableField() {};
	KPGDataTableField(const QString &strOldValue)
	{
		m_strOldValue = strOldValue;
		m_bIsDirty = false;
	}
	
	void setDirty() { m_bIsDirty = true; }
	void clearDirty() { m_bIsDirty = false; }
	
	const QString & oldValue() const { return m_strOldValue; }
	const bool isDirty() const { return m_bIsDirty; }
	
protected:

	QString m_strOldValue; 	// Old value of field, used for cancelChanges()
	bool m_bIsDirty;		// Is field changed by user ?
};

/**
Editable datatable to edit database table content

	@author Lumir Vanek <lvanek@users.sourceforge.net>
*/
class KPGDataTable : public QTable
{
	Q_OBJECT
public:
    KPGDataTable(QWidget *parent, KPGConnection *, const QString&, KPGTable *, const MapIndexKey &, const QString &, const QString &);
    ~KPGDataTable();
   
     // Datatable states
     enum EState { 	
     				stateView = 0, 	// default state
     				stateUpdate, 	// current row is edited (at least one field is dirty), we pending for UPDATE
     				stateInsert, 	// last row is edited, we pending for INSERT 
     				stateRefresh	// refreshing data, waiting for resultset
     			};
     			
     // DML types
     enum EDmlType { 	
     				dmlInsert = 0, 
     				dmlUpdate, 	
     				dmlDelete 
     			};
     			     			
	// List of fields for edited row 	
	typedef QValueList<KPGDataTableField> ListRowFields;
     
    // List of table columns 	
	typedef QValueList<KPGDataTableColumn> ListTableColumns;
     
    // Get actions state
    const DatatableActions & datatableActions() const { return m_datatableActions; }
	
	// Filters events
	bool eventFilter(QObject *, QEvent *);
	
	// Run SQL query for reload data from database table
	void reloadData();
	
	// Terminate thread that run SQL query
	void stopReload();
	
	// Returns TRUE if the connection thread is running; otherwise returns FALSE.
	bool running() const { return  m_ConnectionInThread.running(); }
	
	// Set connection. Don't forget to set 0 if it is disconnected
	void setConnection(KPGConnection* pConnection) 
	{  
	   m_pConnection = pConnection; 
	   if(pConnection == 0)
	   {
	       disableAllActions();
	   }
	}
			
	// Get connection.
	KPGConnection* connection() const { return m_pConnection; }
	
	// Move current row in data table
	void firstRow();
	void previousRow();
	void nextRow();
	void lastRow();
	
	// Edit selected cell in data table
	void editCell();
	
	// Clear selected cell in data table
    void clearCell();
    
    // Insert new row into data table
    void insertRow();
    
    // Delete selected row from data table
    void deleteRow();
    
    // UPDATE or INSERT changes from data table current row to database
    bool commitChanges();
    
    // Cancel changes in data table
    void cancelChanges();
	
	// LOB manipulation functions
	void loadLob(const KURL &);
	void saveLob(const KURL &);
	void editLobInEditor();
	
	// Find first occurence of text
    void findFirst(QStringList &);
    
    // Find next occurence of text
    void findNext();
	
protected:

	// This event handler is called whenever the QScrollView receives a mousePressEvent(): the press position in e is translated to be a point on the contents.
	virtual void contentsMousePressEvent( QMouseEvent* );
	
	// Reimplemented for update actions state purpose
	virtual QWidget *beginEdit(int, int, bool);
    virtual void endEdit(int, int, bool, bool);

    // Disable all actions
    void disableAllActions();
    
    // Enable/disable actions with regards to datatable state
    void setActions();
    
    // Receive event from KPGConnectionInThread
    virtual void customEvent( QCustomEvent * );

    // Display SQL result
    void displayResult();  
    
    // Display data from one reultset row
    void displayOneRow(int, unsigned int, unsigned int, pqxx::result &);
    
    // Refresh current row from database
    void refreshRow();
    
    // Generate new primary key value from sequence
	int generateSerialPrimaryKeyValue(pqxx::work &);
	long generateBigSerialPrimaryKeyValue(pqxx::work &);
    
    // Prepare DML statements
    void prepareInsertStatement(const std::string &);
    void prepareUpdateStatement(const std::string &);
    void prepareDeleteStatement(const std::string &);
    void prepareSelectStatement(const std::string &);
        
    // Build WHERE predicate for primary key columns. 
    const QString wherePredicateForPrimaryKey(int) const;
    
    // Fill key values into vParams, from iParamsIndex position
    void fillParametersForPrimaryKey(int, pqxx::prepare::invocation &) const;
    
    // Run DML string on DB server
	bool runDml(EDmlType);
	
	// Return SQL query for selecting edited row
    const QString sqlForSelectOneRow() const;
	
	// Return SQL query for selecting edited row
    const QString sqlForSelectAllRows() const;
    
    // Retrieve largeobject of LOB for given row and column
    pqxx::largeobject getLargeObject(int, int, std::string &, work &);
    
    // Store largeobject of LOB to database, for given row and column
    bool setLargeObject(pqxx::largeobject &, int, int, std::string &, work &);
      
private:
	
	void fixRow2( int &row, int y );
    
protected:
	// Datatable state
	EState m_state;
	
	// True, if current cell is edited
	bool m_bInplaceEditing;

	// Number of row with edited data
	int m_editedRow;
	
	// List of fields for m_editedRow - that user edited it's value
	ListRowFields m_listRowFields;
		
	// List of columns in table
	ListTableColumns m_listTableColumns;	
		
	// State of actions - enable or disabled
	DatatableActions m_datatableActions;
	    
    // Connection to database
    KPGConnection* m_pConnection;
    
    // Database table to edit its data
    QString m_strTableName;
        
    // Table namespace
    QString m_strNamespaceName;
          	
  	// Map of primary key index columns - pair of <int iColumnNum, QString strColumnName> 
  	MapIndexKey m_mapIndexKey;
  	
  	// Is PK autogenerated using sequence ?
  	bool m_bAutoGeneratedPrimaryKey;
  	
  	// If m_bAutoGeneratedPrimaryKey = true, contain SQL for generate new PK value
  	QString m_strNextValueForPrimaryKey;
    
    // If m_bAutoGeneratedPrimaryKey = true, contain PK column name
    QString m_strPKColumnName;
    
    // Thread that run queries on background
	KPGConnectionInThread m_ConnectionInThread;
	
	// The WHERE predicate, used to reduce anmount of edited data
	QString m_strWherePredicate;
	
	// The ORDER BY predicate, used to sort edited data
	QString m_strOrderByPredicate;
	
	// True, is show DML statements for user inspection, before executing it
	bool m_bShowDml;
	
	// True, if it's first load data 
	bool m_bOpeningReload;
	
	// Resultset to display
	pqxx::result m_pqxxResult;
	
private:
	// A generic implementation of the "find" function
	KFind *m_pFind;
	
	int m_iRowToSearch;
	int m_iColToSearch;
	
protected slots:

    // Called, when the current cell in m_pTableData has changed to row, col
    void slotCurrentChanged(int, int);
    
    // Called, when user changed the value in the m_pTableData
	void slotValueChanged (int, int);
	
	// Called when mouse button button is double-clicked over m_pTableData
	void slotDoubleClicked(int, int, int, const QPoint &);
    
    // Highligth found text
    void slotHighlight( const QString &, int, int);
    
    // Find next occurence of text
    void slotFindNext();
    
signals:
    // Called when SQL query started or finished, or actions need enable/disable
	void sigDataTableStateChanged(KPGDataTable *);
	
	// Emitted, when a context menu is requested 
	void sigContextMenuRequested(const QString & szMenuName, const QPoint & pos );
};

#endif
