//
// C++ Implementation: kpgsqledit
//
// Description: 
//
//
// Author: Lumir Vanek <lvanek@users.sourceforge.net>, (C) 2006
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "kpgsqledit.h"

// include files for KDE
#include <klocale.h>
#include <kdebug.h>
#include <kfinddialog.h>
#include <kreplacedialog.h>
#include <kfind.h>
#include <kreplace.h>

// include files for Qt
#include <private/qrichtext_p.h> // We need QTextCursor

// application specific includes
#include "kpgsyntaxhighlighter.h"
#include "kpgcompletionbox.h"
#include "../kpogreview.h"
#include "../kpgconfiguration.h"
#include "kpgsqleditorsettings.h"

KPGSqlEdit::KPGSqlEdit(QWidget* pParent, const char* szName)
 : KTextEdit(pParent, szName),
 m_pFind(0),
 m_pReplace(0),
 m_bIgnoreSelectionClearing(false)
{
	m_pSyntaxHighlighter = new KPGSyntaxHighlighter(this);
	m_pCompletionBox = new KPGCompletionBox(this);
	
	// Add SQL keywords to list
    QStringList strKeyWords = KPGSyntaxHighlighter::listOfKeyWords();
    
    for(QStringList::const_iterator cit = strKeyWords.begin(); cit != strKeyWords.end(); ++cit)
    {
        // SQL keywords are identified as nodeUnselected
        m_listOfCodeCompletitionObjects.append(KPGOidName(0, *cit, KPGTreeItem::nodeUnselected, KPGTreeItem::m_pIconNull));
    }
	completionObject()->insertItems(KPGSyntaxHighlighter::listOfKeyWords());
		
	// SQL is case insensitive language
	completionObject()->setIgnoreCase(true); 
		
	// Apply current configuration
	slotSqlEditorSettingsChanged();
	
 	// and make sure to be informed about its changes.
	connect(KPoGreView::configuration()->sqleditor(), SIGNAL(sigChanged()), this, SLOT(slotSqlEditorSettingsChanged()));
	connect(m_pCompletionBox, SIGNAL(activated(const QString &)), this, SLOT(setCompletedText(const QString &)));
	connect(this, SIGNAL(textChanged()), this, SLOT(slotTextChanged()));
}


KPGSqlEdit::~KPGSqlEdit()
{
	delete m_pSyntaxHighlighter;
	if(m_pFind != 0) delete m_pFind;
	if(m_pReplace != 0) delete m_pReplace;
}

void KPGSqlEdit::slotSqlEditorSettingsChanged()
{
	m_pSyntaxHighlighter->setColorDefaultText( KPoGreView::configuration()->sqleditor()->colorDefaultText() );
	m_pSyntaxHighlighter->setColorKeyWords( KPoGreView::configuration()->sqleditor()->colorKeyWords() );
	m_pSyntaxHighlighter->setColorDataTypes( KPoGreView::configuration()->sqleditor()->colorDataTypes() );
	m_pSyntaxHighlighter->setColorOperators( KPoGreView::configuration()->sqleditor()->colorOperators() );
	m_pSyntaxHighlighter->setColorQuotedStrings( KPoGreView::configuration()->sqleditor()->colorQuotedStrings() );
	m_pSyntaxHighlighter->setColorNumbers( KPoGreView::configuration()->sqleditor()->colorNumbers() );
    m_pSyntaxHighlighter->setColorComments( KPoGreView::configuration()->sqleditor()->colorComments() );
	
    setFont( KPoGreView::configuration()->sqleditor()->font() );
    
	m_pSyntaxHighlighter->rehighlight();
}

// Filters events - for watching Ctrl+SPACE
bool KPGSqlEdit::eventFilter(QObject *pObject, QEvent *pEvent)
{
	switch(pEvent->type()) 
	{
		case QEvent::KeyPress: 
		{
			QKeyEvent *pKeyEvent = (QKeyEvent*) pEvent;
				
			if((pKeyEvent->key() == Key_Space) && (pKeyEvent->state() & ControlButton) && (m_pCompletionBox->isVisible() == false))
			{
			    updateCtrlSpaceCompletition(true); 
                return true;
			}
			
			break;
		}
		default: break;
	}
	
	return KTextEdit::eventFilter(pObject, pEvent);
}

// Update completition when taxt changed
void KPGSqlEdit::slotTextChanged()
{
    if(m_pCompletionBox->isVisible())
	{
		updateCtrlSpaceCompletition(false);
	}
}

const QString KPGSqlEdit::uncompletedWord()
{
	// Get cursor position 
	int para, index;
	getCursorPosition(&para, &index );								
												
	// Get text at cursor position, look at last word before cursor
	QString strText(text(para));
	QString strOneWord("");
	for(unsigned int i = 0; i < strText.length(); i++)
	{
		if(i >= (unsigned int) index) break;
					
		if(strText[i] == ' ')
		{
		  strOneWord.truncate(0);
		}
		else
		{
		  strOneWord.append(strText[i]);
		}	
	}
	
	return strOneWord;
}

// Get uncompleted word, set it to completition and set completed items to box, eventually show it
void KPGSqlEdit::updateCtrlSpaceCompletition(bool bShow)
{
    QString strWord(uncompletedWord());
	kdDebug() << strWord << endl;
												
	// Set-up and show completion box
	QStringList strListMatches;
	if(strWord.length() > 0)
	{											
		strListMatches = completionObject()->allMatches(strWord.upper());
	}	
	else
	{
	    for(KPGOidNameList::const_iterator cit = m_listOfCodeCompletitionObjects.begin(); cit != m_listOfCodeCompletitionObjects.end(); ++cit)
	    {
            strListMatches.append((*cit).name());
        }
	}	
		
	if(strListMatches.count() > 0)
	{
		QTextCursor *pTextCursor = textCursor();
												
		// Get cursor place, make correction to display completion box on proper place
		QPoint point(pTextCursor->globalX(), pTextCursor->globalY() + 18); // TODO: 18 -> fontHigth
		point = viewport()->mapToGlobal(contentsToViewport(point));
										
		setCompletedItems(strListMatches);
		m_pCompletionBox->move(point);
			
		if(bShow)
		    m_pCompletionBox->show();
	}
	else
	{
	    if(m_pCompletionBox->isVisible())
	    {
	        m_pCompletionBox->hide();
	    }
	}
}

void KPGSqlEdit::setCompletionObject(KCompletion* comp, bool hsig)
{
    KCompletion *oldComp = compObj();
    if( oldComp && handleSignals() )
       disconnect( oldComp, SIGNAL( matches( const QStringList& )), this, SLOT( setCompletedItems( const QStringList& )));

    if( comp && hsig )
      connect( comp, SIGNAL( matches( const QStringList& )), this, SLOT( setCompletedItems( const QStringList& )));

    KCompletionBase::setCompletionObject( comp, hsig );
}

// Insert code completion
void KPGSqlEdit::setCompletedText(const QString& strCompletion)
{
	// Get cursot position 
	int para, index;
	getCursorPosition(&para, &index);								
												
	// Get text at cursor position, look at last word before cursor
	QString strText = text(para);
	for(int i = index; i > 0; i--)
	{
		if(((strText[i-1] == ' ') || (strText[i-1] == '\t') || (strText[i-1] == '\n')) /*&& (i < index)*/) break;
		//kdDebug() << "Removing: " << strText[i-1] << endl;	
		doKeyboardAction(ActionBackspace); // remove uncomplete word
	}
	
	getCursorPosition(&para, &index );	
	insertAt(strCompletion, para, index ); // insert completed word
	KTextEdit::setCursorPosition(para, index +  strCompletion.length());	
}

void KPGSqlEdit::setCompletedItems(const QStringList& items)
{
    QString txt;
    if (m_pCompletionBox && m_pCompletionBox->isVisible() ) 
    {
        // The popup is visible already - do the matching on the initial string,
        // not on the currently selected one.
        txt = m_pCompletionBox->cancelledText();
    } 
    else 
    {
        txt = uncompletedWord();
    }
    
    if ( !items.isEmpty() && !(items.count() == 1 && txt == items.first()) )
    {
    	if ( m_pCompletionBox->isVisible() )
        {
        	bool wasSelected = m_pCompletionBox->isSelected( m_pCompletionBox->currentItem() );
            const QString currentSelection = m_pCompletionBox->currentText();
            
            m_pCompletionBox->setItems(items, m_listOfCodeCompletitionObjects);
            
            QListBoxItem* item = m_pCompletionBox->findItem( currentSelection, Qt::ExactMatch );
            // If no item is selected, that means the listbox hasn't been manipulated by the user yet,
            // because it's not possible otherwise to have no selected item. In such case make
            // always the first item current and unselected, so that the current item doesn't jump.
            if( !item || !wasSelected )
            {
                wasSelected = false;
                item = m_pCompletionBox->item( 0 );
            }
            if ( item )
            {
                m_pCompletionBox->blockSignals( true );
                m_pCompletionBox->setCurrentItem( item );
                m_pCompletionBox->setSelected( item, wasSelected );
                m_pCompletionBox->blockSignals( false );
            }
        }
        else // completion box not visible yet -> show it
        {
            if ( !txt.isEmpty() )
	            m_pCompletionBox->setCancelledText( txt );
            
            // TODO: rewrote it to insert text + icon for completion items !!!
            m_pCompletionBox->setItems(items, m_listOfCodeCompletitionObjects);
            m_pCompletionBox->popup();
        }
    }
    else
    {
        if(m_pCompletionBox && m_pCompletionBox->isVisible())
            m_pCompletionBox->hide();
    }
}

// Set list of DB objects for code completition
void KPGSqlEdit::setListOfObjectsForCodeCompletition(const KPGOidNameList &listOfCodeCompletitionObjects)
{
    clearListOfObjectsForCodeCompletition();
    
    // Add given DB objects
    for(KPGOidNameList::const_iterator cit = listOfCodeCompletitionObjects.begin(); cit != listOfCodeCompletitionObjects.end(); ++cit)
    {
        m_listOfCodeCompletitionObjects.append(*cit);
    }
    
    // Sort list of code completitions
    m_listOfCodeCompletitionObjects.sort();
    
    // Create list of all completions
    QStringList strAllCompletions;
    for(KPGOidNameList::const_iterator cit = m_listOfCodeCompletitionObjects.begin(); cit != m_listOfCodeCompletitionObjects.end(); ++cit)
	{
        strAllCompletions.append((*cit).name());
    }
        
    completionObject()->insertItems(strAllCompletions);
}

// Clear list of DB objects for code completition
void KPGSqlEdit::clearListOfObjectsForCodeCompletition()
{
	m_listOfCodeCompletitionObjects.clear();
    completionObject()->clear();
    
    // Add SQL keywords to list
    QStringList strKeyWords = KPGSyntaxHighlighter::listOfKeyWords();
    
    for(QStringList::const_iterator cit = strKeyWords.begin(); cit != strKeyWords.end(); ++cit)
    {
        // SQL keywords are identified as nodeUnselected
        m_listOfCodeCompletitionObjects.append(KPGOidName(0, *cit, KPGTreeItem::nodeUnselected, KPGTreeItem::m_pIconNull));
    }
}

// Set cursor position to given character offset, select character
void KPGSqlEdit::highlightCharacter(int iPosition)
{
	for(int iParagraph = 0; iParagraph < paragraphs(); iParagraph++)
	{
		if(iPosition > paragraphLength(iParagraph))
		{
			iPosition -= paragraphLength(iParagraph) + 1;
		}
		else
		{
			KTextEdit::setCursorPosition(iParagraph, iPosition);
			KTextEdit::setSelection(iParagraph, iPosition-1, iParagraph, iPosition);
			KTextEdit::ensureCursorVisible();
			return;
		}
	}
}

// Find first occurence of text
void KPGSqlEdit::findFirst(QStringList &listOfSearchHistory)
{
    // don't listen to selection changes
    m_bIgnoreSelectionClearing = true;
    
    KFindDialog dlg(this, "", 0, listOfSearchHistory, hasSelectedText());
    dlg.setHasCursor(false);
        
    int c = dlg.exec();
    // listen to selection changes
    m_bIgnoreSelectionClearing = false;
    
    if(c != QDialog::Accepted)
        return;

    listOfSearchHistory = dlg.findHistory();
    
    if(m_pFind != 0) delete m_pFind;
    m_pFind = new KFind(dlg.pattern(), dlg.options(), this);

    // Connect highlight signal to code which handles highlighting
    // of found text.
    connect( m_pFind, SIGNAL( highlight( const QString &, int, int ) ),
             this, SLOT( slotHighlight( const QString &, int, int ) ) );
             
    // Connect findNext signal - called when pressing the button in the dialog
    connect( m_pFind, SIGNAL( findNext() ), this, SLOT( slotFindNext() ) );

	setUpFindBoundaries(dlg.options(), &m_iStartPara, &m_iStartIdx, &m_iEndPara, &m_iEndIdx);
    
    m_iParaToSearch = m_iStartPara;
    m_pFind->setData(text(m_iStartPara), m_iStartIdx);

    slotFindNext();
}

// Find next occurence of text
void KPGSqlEdit::findNext()
{
	if (!m_pFind) // shouldn't be called before find is activated
        return;
        
	getCursorPosition(&m_iParaToSearch, &m_iStartIdx);
	m_pFind->setData(text(m_iParaToSearch), m_iStartIdx);
	
	slotFindNext();
}

// Find next occurence of text
void KPGSqlEdit::slotFindNext() 
{
    if (!m_pFind) // shouldn't be called before find is activated
        return;
	
    KFind::Result res = KFind::NoMatch;
    while ( res == KFind::NoMatch &&
            m_iParaToSearch <= QMAX(m_iStartPara, m_iEndPara) &&
            m_iParaToSearch >= QMIN(m_iStartPara, m_iEndPara) 
          ) 
    	{
    		//kdDebug() << "searching  1:" << m_iParaToSearch << " 2: " << m_iEndPara << " 3: " << m_iStartPara << endl;
        
        	if(m_pFind->needData()) 
        	{
            	int idx = (m_iParaToSearch == m_iEndPara) ? m_iEndIdx : -1;
            	m_pFind->setData(text(m_iParaToSearch), idx);
        	}
        
        	// Let KFind inspect the text fragment, and display a dialog if a match is found
        	res = m_pFind->find();
        
        	if( res == KFind::NoMatch ) 
        	{
            	m_iParaToSearch += (m_pFind->options() & KFindDialog::FindBackwards) ? -1 : 1;
        	}
        	//kdDebug() << "2searching  1:" << m_iParaToSearch << " 2: " << m_iEndPara << " 3: " << m_iStartPara << endl;
		}

    if( res == KFind::NoMatch ) 
    { // i.e. at end
        m_pFind->displayFinalDialog();
        m_pFind->resetCounts();
        //delete m_pFind;
        //m_pFind = 0L;
        removeSelection(0);
    }
}

// Highligth found text
void KPGSqlEdit::slotHighlight( const QString &, int index, int length)
{
    //kdDebug() << "highlight: " << index << " " << length << endl;
    setSelection(m_iParaToSearch, index, m_iParaToSearch, index + length );
}

void KPGSqlEdit::selectAll( bool b )
{
    if (!b && m_bIgnoreSelectionClearing)
        return;
    
    KTextEdit::selectAll(b);
}

// Replace first occurence of text
void KPGSqlEdit::replace(QStringList &listOfSearchHistory, QStringList &listOfReplacementHistory)
{
	// don't listen to selection changes
    m_bIgnoreSelectionClearing = true;
    
    KReplaceDialog dlg(this, "", 0, listOfSearchHistory, listOfReplacementHistory, hasSelectedText());
    dlg.setHasCursor(false);
        
    int c = dlg.exec();
    // listen to selection changes
    m_bIgnoreSelectionClearing = false;
    
    if(c != QDialog::Accepted)
        return;

    listOfSearchHistory = dlg.findHistory();
    listOfReplacementHistory = dlg.replacementHistory();
        
    
    if(m_pReplace != 0) delete m_pReplace;
    m_pReplace = new KReplace(dlg.pattern(), dlg.replacement(), dlg.options(), this);

    // Connect highlight signal to code which handles highlighting
    // of found text.
    connect( m_pReplace, SIGNAL( highlight( const QString &, int, int ) ),
             this, SLOT( slotHighlight( const QString &, int, int ) ) );
             
    // Connect findNext signal - called when pressing the button in the dialog
    connect( m_pReplace, SIGNAL( findNext() ), this, SLOT( slotReplaceNext() ) );

	// Connect replace signal - called when doing a replacement
  	connect( m_pReplace, SIGNAL( replace(const QString &, int, int, int) ),
          this, SLOT( slotReplace(const QString &, int, int, int) ) );

	setUpFindBoundaries(dlg.options(), &m_iStartPara, &m_iStartIdx, &m_iEndPara, &m_iEndIdx);
	
    m_iParaToSearch = m_iStartPara;
    m_pReplace->setData(text(m_iStartPara ), m_iStartIdx);

    slotReplaceNext();
}

// Replace next occurence of text
void KPGSqlEdit::slotReplaceNext()
{
    if (!m_pReplace) // shouldn't be called before find is activated
        return;

	KFind::Result res = KFind::NoMatch;
    while ( res == KFind::NoMatch &&
            m_iParaToSearch <= QMAX(m_iStartPara, m_iEndPara ) &&
            m_iParaToSearch >= QMIN(m_iStartPara, m_iEndPara ) ) 
    	{
    		//kdDebug() << "searching  1:" << m_iParaToSearch << " 2: " << m_iEndPara << " 3: " << m_iStartPara << endl;
        
        	if(m_pReplace->needData()) 
        	{
            	int idx = (m_iParaToSearch == m_iEndPara) ? m_iEndIdx : -1;
            	m_pReplace->setData( text(m_iParaToSearch), idx);
        	}
        
        	// Let KFind inspect the text fragment, and display a dialog if a match is found
        	res = m_pReplace->replace();
        
        	if( res == KFind::NoMatch ) 
        	{
            	m_iParaToSearch += (m_pReplace->options() & KFindDialog::FindBackwards) ? -1 : 1;
        	}
        	//kdDebug() << "2searching  1:" << m_iParaToSearch << " 2: " << m_iEndPara << " 3: " << m_iStartPara << endl;
		}

    if( res == KFind::NoMatch ) 
    { // i.e. at end
        m_pReplace->displayFinalDialog();
        delete m_pReplace;
        m_pReplace = 0L;
        removeSelection(0);
    }
    
    m_pSyntaxHighlighter->rehighlight();
}

// Implement updating of replaced text during the replace operation
void KPGSqlEdit::slotReplace(const QString &strText, int replacementIndex, int replacedLength, int /*matchedLength*/)
{
	kdDebug() << k_funcinfo << " " << strText << " replacementIndex: " << replacementIndex << " replacedLength: " << replacedLength<< endl;
	
	// don't listen to selection changes
    m_bIgnoreSelectionClearing = true;
    m_pSyntaxHighlighter->disable();
    
	removeParagraph(m_iParaToSearch);
	insertParagraph(strText, m_iParaToSearch);
	
	// listen to selection changes
    m_bIgnoreSelectionClearing = false;
    m_pSyntaxHighlighter->enable();
}
  
// Set up boudaries for find/replace
void KPGSqlEdit::setUpFindBoundaries(long lOptions, int *piStartPara, int *piStartIdx, int *piEndPara, int *piEndIdx)
{ 
    if(lOptions & KFindDialog::SelectedText ) 
    {
        Q_ASSERT(hasSelectedText());
        if(lOptions & KFindDialog::FindBackwards) 
        	getSelection(piEndPara, piEndIdx, piStartPara, piStartIdx);
        else
        	getSelection(piStartPara, piStartIdx, piEndPara, piEndIdx);
    } 
    else 
    {
    	*piStartIdx  = -1; // Interpreted as "process all data" by KFind
    	*piEndIdx    = -1; // Interpreted as "process all data" by KFind
    	
    	if(lOptions & KFindDialog::FindBackwards) 
		{
			*piStartPara = paragraphs();
			*piEndPara   = 0;
		}
    	else
    	{
        	*piStartPara = 0;
        	*piEndPara   = paragraphs();
        }
    }
}

void KPGSqlEdit::virtual_hook( int id, void* data )
{ KCompletionBase::virtual_hook( id, data ); }
